def connect(self, key):
    import fabric.network
    import fabric.state
    r = fabric.network.parse_host_string(key)
    user = r['user'] or 'root'
    host = r['host']
    server = fabric.state.env.instances[host]
    try:
        ssh_info = server.init_ssh_key(user=user)
    except server.paramiko.SSHException as e:
        from ploy_fabric import log
        import sys
        log.error("Couldn't validate fingerprint for ssh connection.")
        log.error(e)
        log.error("Is the server finished starting up?")
        sys.exit(1)
    self[key] = ssh_info['client']


def patch():
    import fabric.network
    fabric.network.HostConnectionCache.connect = connect



from contextlib import contextmanager
from functools import wraps
import argparse
import inspect
import logging
import os
import sys


log = logging.getLogger('ploy_fabric')


notset = object()


def get_host_string(instance, user=None):
    return "{user}@{host}".format(
        user=instance.config.get('user', user or 'root'),
        host=instance.uid)


class context:
    def __init__(self, *args, **kwargs):
        self.user = kwargs.get('user')
        if args:
            self(args)

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            import fabric.state
            with fabric.state.env.instance.fabric(user=self.user):
                return func(*args, **kwargs)
        return wrapper


@contextmanager
def instance_context(instance, user=None):
    with fabric_env(instance.master.ctrl, instance) as env:
        env.host_string = get_host_string(instance, user=user)
        yield


def has_fabfile(instance):
    fabfile = instance.config.get('fabfile')
    if fabfile is None:
        return False
    if not os.path.exists(fabfile):
        return False
    return True


def get_fabfile(instance):
    fabfile = instance.config.get('fabfile')
    if fabfile is None:
        log.error("No fabfile declared.")
        sys.exit(1)
    if not os.path.exists(fabfile):
        log.error("The fabfile at '%s' is missing." % fabfile)
        sys.exit(1)
    return fabfile


@contextmanager
def sys_argv(newargv):
    orig_sys_argv = sys.argv
    sys.argv = newargv
    try:
        yield
    finally:
        sys.argv = orig_sys_argv


@contextmanager
def cwd(path):
    orig_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(orig_cwd)


class DeprecationProxy:
    def __init__(self, wrapped, name, new_name):
        self.wrapped = wrapped
        self.name = name
        self.new_name = new_name
        self.seen = set()

    def __getattr__(self, name):
        caller_frame = inspect.currentframe().f_back
        info = inspect.getframeinfo(caller_frame)
        orig_name = self.__dict__['name']
        new_name = self.__dict__['new_name']
        seen = self.__dict__['seen']
        key = (info.filename, info.lineno)
        if key not in seen:
            seen.add(key)
            log.warning("Use of deprecated variable name '%s', use '%s' instead.\n%s:%s\n%s" % (
                orig_name, new_name, info.filename, info.lineno, ''.join(info.code_context)))
        return getattr(self.__dict__['wrapped'], name)


def get_fabric_env_settings(ctrl, instance):
    env = dict(
        reject_unknown_hosts=True,
        disable_known_hosts=True,
        host_string=get_host_string(instance),
        known_hosts=ctrl.known_hosts,
        instances=ctrl.instances,
        instance=instance,
        config_base=ctrl.config.path)
    for k, v in instance.config.items():
        if not k.startswith('fabric-'):
            continue
        env[k[7:]] = v
    return env


@contextmanager
def fabric_env(ctrl, instance, fabcmd=False):
    import fabric.state
    env, env_options = fabric.state.env, fabric.state.env_options
    orig = dict()
    orig_options = list(env_options)
    for k, v in get_fabric_env_settings(ctrl, instance).items():
        orig[k] = env.get(k, notset)
        env[k] = v
    new_options = []
    for option in env_options:
        if option.dest not in orig:
            new_options.append(option)
        elif fabcmd:
            log.warn("Removed fab command line option %s because of overwrite for instance %s." % (option, instance.config_id))
    env_options[:] = new_options
    for new_k, old_k in (('instances', 'servers'), ('instance', 'server')):
        orig[old_k] = env.get(old_k, notset)
        env[old_k] = DeprecationProxy(orig[new_k], old_k, new_k)
    try:
        yield env
    finally:
        for key, value in orig.items():
            if value is notset:
                if key in env:
                    del env[key]
            else:
                env[key] = value
        env_options[:] = orig_options


@contextmanager
def fabric_connections(ctrl, instance, fabcmd=False):
    import fabric.state
    original_connections = set(fabric.state.connections)
    try:
        with fabric_env(ctrl, instance, fabcmd=fabcmd):
            yield
    finally:
        for connection in set(fabric.state.connections) - original_connections:
            log.debug("Closing connection to '%s' after Fabric call." % connection)
            del fabric.state.connections[connection]


@contextmanager
def fabric_integration(ctrl, instance, fabcmd=False):
    from ploy_fabric import _fabric_integration
    # this needs to be done before any other fabric module import
    _fabric_integration.patch()

    with cwd(os.path.dirname(get_fabfile(instance))):
        with fabric_connections(ctrl, instance, fabcmd=fabcmd):
            yield


class FabricCmd(object):
    """Run Fabric"""

    def __init__(self, ctrl):
        self.ctrl = ctrl

    def __call__(self, argv, help):
        """Do stuff on the cluster (using fabric)"""
        parser = argparse.ArgumentParser(
            prog="%s fab" % self.ctrl.progname,
            description=help,
            add_help=False,
        )
        instances = self.ctrl.get_instances(command='init_ssh_key')
        parser.add_argument("instance", nargs=1,
                            metavar="instance",
                            help="Name of the instance from the config.",
                            choices=list(instances))
        parser.add_argument("fabric_opts",
                            metavar="...", nargs=argparse.REMAINDER,
                            help="Fabric options")
        args = parser.parse_args(argv)

        instance = instances[args.instance[0]]
        with fabric_integration(self.ctrl, instance, fabcmd=True):
            from fabric.main import main
            fabfile = get_fabfile(instance)
            newargv = ['fab', '-f', fabfile]
            if args.fabric_opts:
                newargv = newargv + args.fabric_opts
            with sys_argv(newargv):
                main()


class DoCmd(object):
    """Run a specific Fabric task"""

    def __init__(self, ctrl):
        self.ctrl = ctrl

    def get_completion(self):
        instances = set()
        for instance in self.ctrl.get_instances(command='do'):
            if self.ctrl.instances[instance].has_fabfile():
                instances.add(instance)
        return sorted(instances)

    def __call__(self, argv, help):
        """Run a fabric task on an instance"""
        parser = argparse.ArgumentParser(
            prog="%s do" % self.ctrl.progname,
            description=help)
        parser.add_argument("instance", nargs=1,
                            metavar="instance",
                            help="Name of the instance from the config.",
                            choices=self.get_completion())
        group = parser.add_mutually_exclusive_group(required=True)
        group.add_argument("task", nargs='?',
                           help="The task to run.")
        group.add_argument("-l", "--list",
                           action='store_true',
                           help="List available tasks.")
        parser.add_argument("task_args",
                            metavar="arg|key=value", nargs="*",
                            help="Arguments for the task.")
        args = parser.parse_args(argv)

        instance = self.ctrl.instances[args.instance[0]]
        if args.list:
            print "Available commands:"
            print
            with callables(instance) as tasks:
                for name in sorted(tasks):
                    print "    %s" % name
                return
        task_args = []
        task_kwargs = {}
        for arg in args.task_args:
            parts = arg.split('=', 1)
            if len(parts) == 1:
                task_args.append(arg)
            else:
                task_kwargs[parts[0]] = parts[1]
        instance.do(args.task, *task_args, **task_kwargs)


@contextmanager
def callables(instance):
    with fabric_integration(instance.master.ctrl, instance):
        from fabric.main import extract_tasks
        from fabric.state import env

        fabfile = get_fabfile(instance)
        with open(fabfile) as f:
            source = f.read()
        code = compile(source, fabfile, 'exec')
        g = {'__file__': fabfile}
        exec code in g, g
        new_style, classic, default = extract_tasks(g.items())
        callables = new_style if env.new_style_tasks else classic
        yield callables


def do(self, task, *args, **kwargs):
    with callables(self) as tasks:
        return tasks[task](*args, **kwargs)


def augment_instance(instance):
    if not hasattr(instance, 'init_ssh_key'):
        return
    if not hasattr(instance, 'fabric'):
        instance.fabric = instance_context.__get__(instance, instance.__class__)
    if not hasattr(instance, 'has_fabfile'):
        instance.has_fabfile = has_fabfile.__get__(instance, instance.__class__)
    if not hasattr(instance, 'do'):
        instance.do = do.__get__(instance, instance.__class__)


def get_commands(ctrl):
    return [
        ('fab', FabricCmd(ctrl)),
        ('do', DoCmd(ctrl))]


def get_massagers():
    from ploy.config import PathMassager
    return [PathMassager(None, 'fabfile')]


plugin = dict(
    augment_instance=augment_instance,
    get_commands=get_commands,
    get_massagers=get_massagers)



from mock import MagicMock, patch
from ploy import Controller
import logging
import pytest


def caplog_messages(caplog, level=logging.INFO):
    return [
        x.message
        for x in caplog.records
        if x.levelno >= level]


@pytest.fixture
def ctrl(ployconf):
    import ploy_fabric
    import ploy.tests.dummy_plugin
    ctrl = Controller(ployconf.directory)
    ctrl.plugins = {
        'dummy': ploy.tests.dummy_plugin.plugin,
        'fabric': ploy_fabric.plugin}
    ctrl.configfile = ployconf.path
    return ctrl


@pytest.yield_fixture
def fabfile(tempdir):
    import sys
    sys.modules.pop('foo', None)
    yield tempdir['etc/foo.py']
    sys.modules.pop('foo', None)


@pytest.mark.parametrize("cmd_name", ['fab', 'do'])
@pytest.fixture(params=['do', 'fab'])
def cmd(request):
    return request.param


@pytest.fixture
def open_session_mock(monkeypatch):
    import ploy.tests.dummy_plugin
    open_session_mock = MagicMock()
    open_session_mock().recv.return_value = ''
    open_session_mock().recv_stderr.return_value = ''
    open_session_mock().exit_status_ready.return_value = True
    open_session_mock().recv_exit_status.return_value = 0
    monkeypatch.setattr(
        ploy.tests.dummy_plugin.MockTransport, 'open_session',
        open_session_mock, raising=False)
    return open_session_mock


class TestDoCommand:
    def testCallWithExistingInstanceButTooViewArguments(self, ctrl, fabfile, ployconf):
        ployconf.fill([
            '[dummy-instance:foo]',
            'host = localhost',
            'fabfile = %s' % fabfile.path])
        fabfile.fill([
            'from fabric.api import run',
            'def something():',
            '    run("something")'])
        with patch('sys.stderr') as StdErrMock:
            with pytest.raises(SystemExit):
                ctrl(['./bin/ploy', 'do', 'foo'])
        output = "".join(x[0][0] for x in StdErrMock.write.call_args_list)
        assert 'one of the arguments' in output

    def testCallWithMissingFabfileDeclaration(self, ctrl, ployconf):
        ployconf.fill([
            '[dummy-instance:foo]'])
        with patch('sys.stderr') as StdErrMock:
            with pytest.raises(SystemExit):
                ctrl(['./bin/ploy', 'do', 'foo', 'something'])
        output = "".join(x[0][0] for x in StdErrMock.write.call_args_list)
        assert "invalid choice: 'foo' (choose from )" in output

    def testCallWithExistingInstance(self, ctrl, fabfile, open_session_mock, ployconf):
        ployconf.fill([
            '[dummy-instance:foo]',
            'host = localhost',
            'fabfile = %s' % fabfile.path])
        fabfile.fill([
            'from fabric.api import run',
            'def something():',
            '    run("something")'])
        ctrl(['./bin/ploy', 'do', 'foo', 'something'])
        assert open_session_mock().recv.called is True
        from fabric.state import env
        assert 'instances' not in env

    def testCallWithTaskArg(self, ctrl, fabfile, ployconf):
        ployconf.fill([
            '[dummy-instance:foo]',
            'host = localhost',
            'fabfile = %s' % fabfile.path])
        fabfile.fill([
            'from fabric.api import run',
            'def something(fooarg):',
            '    print fooarg'])
        with patch('sys.stdout') as StdOutMock:
            ctrl(['./bin/ploy', 'do', 'foo', 'something', 'bararg'])
        output = "".join(x[0][0] for x in StdOutMock.write.call_args_list)
        assert 'bararg' in output

    def testCallWithTaskKwArg(self, ctrl, fabfile, ployconf):
        ployconf.fill([
            '[dummy-instance:foo]',
            'host = localhost',
            'fabfile = %s' % fabfile.path])
        fabfile.fill([
            'from fabric.api import run',
            'def something(fooarg="foo"):',
            '    print fooarg'])
        with patch('sys.stdout') as StdOutMock:
            ctrl(['./bin/ploy', 'do', 'foo', 'something', 'fooarg=bararg'])
        output = "".join(x[0][0] for x in StdOutMock.write.call_args_list)
        assert 'bararg' in output

    def testDeprecation(self, caplog, ctrl, fabfile, ployconf):
        ployconf.fill([
            '[dummy-instance:foo]',
            'host = localhost',
            'fabfile = %s' % fabfile.path])
        fabfile.fill([
            'from fabric.api import env',
            'def something(fooarg="foo"):',
            '    print env.servers',
            '    print env.server'])
        ctrl(['./bin/ploy', 'do', 'foo', 'something'])
        servers_msg, server_msg = [x.splitlines() for x in caplog_messages(caplog, level=logging.WARN)]
        assert servers_msg[0] == "Use of deprecated variable name 'servers', use 'instances' instead."
        parts = servers_msg[1].rsplit(':', 1)
        assert parts[0].endswith('etc/foo.py')
        assert parts[1] == '3'
        assert servers_msg[2] == "    print env.servers"
        assert server_msg[0] == "Use of deprecated variable name 'server', use 'instance' instead."
        parts = server_msg[1].rsplit(':', 1)
        assert parts[0].endswith('etc/foo.py')
        assert parts[1] == '4'
        assert server_msg[2] == "    print env.server"


class TestFabCommand:
    def testCallWithExistingInstanceButTooViewArguments(self, ctrl, ployconf):
        ployconf.fill([
            '[dummy-instance:foo]'])
        with patch('ploy_fabric.log') as LogMock:
            with pytest.raises(SystemExit):
                ctrl(['./bin/ploy', 'fab', 'foo'])
        LogMock.error.assert_called_with('No fabfile declared.')

    def testCallWithMissingFabfileDeclaration(self, ctrl, ployconf):
        ployconf.fill([
            '[dummy-instance:foo]'])
        with patch('ploy_fabric.log') as LogMock:
            with pytest.raises(SystemExit):
                ctrl(['./bin/ploy', 'fab', 'foo', 'something'])
        LogMock.error.assert_called_with('No fabfile declared.')

    def testCallWithExistingInstance(self, ctrl, fabfile, open_session_mock, ployconf):
        ployconf.fill([
            '[dummy-instance:foo]',
            'host = localhost',
            'fabfile = %s' % fabfile.path])
        fabfile.fill([
            'from fabric.api import run',
            'def something():',
            '    run("something")'])
        with pytest.raises(SystemExit) as e:
            ctrl(['./bin/ploy', 'fab', 'foo', 'something'])
        assert open_session_mock().recv.called is True
        assert e.value.code == 0
        from fabric.state import env
        assert 'instances' not in env


def test_call_with_no_arguments(cmd, ctrl, ployconf):
    ployconf.fill('')
    with patch('sys.stderr') as StdErrMock:
        with pytest.raises(SystemExit):
            ctrl(['./bin/ploy', cmd])
    output = "".join(x[0][0] for x in StdErrMock.write.call_args_list)
    assert 'usage: ploy %s' % cmd in output
    assert 'too few arguments' in output


def test_call_with_non_existing_instance(cmd, ctrl, ployconf):
    ployconf.fill('')
    with patch('sys.stderr') as StdErrMock:
        with pytest.raises(SystemExit):
            ctrl(['./bin/ploy', cmd, 'foo'])
    output = "".join(x[0][0] for x in StdErrMock.write.call_args_list)
    assert 'usage: ploy %s' % cmd in output
    assert "argument instance: invalid choice: 'foo'" in output


def test_list_tasks(cmd, ctrl, fabfile, ployconf):
    ployconf.fill([
        '[dummy-instance:foo]',
        'host = localhost',
        'fabfile = %s' % fabfile.path])
    fabfile.fill([
        'def something():',
        '    pass',
        '',
        'def something_else():',
        '    pass',
        ''])
    with patch('sys.stdout') as StdOutMock:
        StdOutMock.isatty.return_value = False
        try:
            ctrl(['./bin/ploy', cmd, 'foo', '-l'])
        except SystemExit as e:
            assert e.code == 0
    output = "".join(x[0][0] for x in StdOutMock.write.call_args_list)
    assert 'Available commands' in output
    assert 'something' in output
    assert 'something_else' in output


def test_env_overwrite_from_config(caplog, cmd, ctrl, fabfile, ployconf):
    ployconf.fill([
        '[dummy-instance:foo]',
        'host = localhost',
        'fabfile = %s' % fabfile.path,
        'fabric-ham = egg',
        'fabric-shell = fooshell'])
    fabfile.fill([
        'from fabric.api import env',
        'def something():',
        '    print "ham", env.ham',
        '    print "shell", env.shell'])
    with patch('sys.stdout') as StdOutMock:
        StdOutMock.isatty.return_value = False
        try:
            ctrl(['./bin/ploy', cmd, 'foo', 'something'])
        except SystemExit as e:
            assert e.code == 0
    output = "".join(x[0][0] for x in StdOutMock.write.call_args_list)
    assert "ham egg\n" in output
    assert "shell fooshell\n" in output
    if cmd == 'fab':
        msgs = [x for x in caplog_messages(caplog, level=logging.WARN)]
        assert msgs == [
            'Removed fab command line option -D/--disable-known-hosts because of overwrite for instance dummy-instance:foo.',
            'Removed fab command line option -r/--reject-unknown-hosts because of overwrite for instance dummy-instance:foo.',
            'Removed fab command line option -s/--shell because of overwrite for instance dummy-instance:foo.']



from setuptools import setup
import os


here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()
HISTORY = open(os.path.join(here, 'HISTORY.rst')).read()


version = "1.1.1"


install_requires = [
    'setuptools',
    'ploy >= 1.0.0, < 2dev',
    'Fabric>=1.4.0,!=1.8.3,<2dev']


setup(
    version=version,
    description="Plugin to integrate Fabric with ploy.",
    long_description=README + "\n\n" + HISTORY,
    name="ploy_fabric",
    author='Florian Schulze',
    author_email='florian.schulze@gmx.net',
    license="BSD 3-Clause License",
    url='http://github.com/ployground/ploy_fabric',
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 2 :: Only',
        'Topic :: System :: Installation/Setup',
        'Topic :: System :: Systems Administration'],
    include_package_data=True,
    zip_safe=False,
    packages=['ploy_fabric'],
    install_requires=install_requires,
    entry_points="""
        [ploy.plugins]
        fabric = ploy_fabric:plugin
    """)
